@@ -27,6 +27,7 @@ import time
 from urllib import parse
 import uuid
 
+from amqp import exceptions as amqp_exec
 import kombu
 import kombu.connection
 import kombu.entity
@@ -87,14 +88,16 @@ rabbit_opts = [
                help='SSL certification authority file '
                     '(valid only if SSL enabled).'),
     cfg.BoolOpt('heartbeat_in_pthread',
-                default=False,
-                help="EXPERIMENTAL: Run the health check heartbeat thread "
-                     "through a native python thread. By default if this "
-                     "option isn't provided the  health check heartbeat will "
-                     "inherit the execution model from the parent process. By "
-                     "example if the parent process have monkey patched the "
+                default=True,
+                help="Run the health check heartbeat thread "
+                     "through a native python thread by default. If this "
+                     "option is equal to False then the health check "
+                     "heartbeat will inherit the execution model "
+                     "from the parent process. For "
+                     "example if the parent process has monkey patched the "
                      "stdlib by using eventlet/greenlet then the heartbeat "
-                     "will be run through a green thread."),
+                     "will be run through a green thread.",
+                deprecated_for_removal=True),
     cfg.FloatOpt('kombu_reconnect_delay',
                  default=1.0,
                  deprecated_group='DEFAULT',
@@ -172,6 +175,11 @@ rabbit_opts = [
                'for direct send. The direct send is used as reply, '
                'so the MessageUndeliverable exception is raised '
                'in case the client queue does not exist.'),
+    cfg.BoolOpt('enable_cancel_on_failover',
+                default=False,
+                help="Enable x-cancel-on-ha-failover flag so that "
+                     "rabbitmq server will cancel and notify consumers"
+                     "when queue is down")
 ]
 
 LOG = logging.getLogger(__name__)
@@ -233,7 +241,8 @@ class Consumer(object):
 
     def __init__(self, exchange_name, queue_name, routing_key, type, durable,
                  exchange_auto_delete, queue_auto_delete, callback,
-                 nowait=False, rabbit_ha_queues=None, rabbit_queue_ttl=0):
+                 nowait=False, rabbit_ha_queues=None, rabbit_queue_ttl=0,
+                 enable_cancel_on_failover=False):
         """Init the Consumer class with the exchange_name, routing_key,
         type, durable auto_delete
         """
@@ -255,10 +264,16 @@ class Consumer(object):
             type=type,
             durable=self.durable,
             auto_delete=self.exchange_auto_delete)
+        self.enable_cancel_on_failover = enable_cancel_on_failover
 
     def declare(self, conn):
         """Re-declare the queue after a rabbit (re)connect."""
 
+        consumer_arguments = None
+        if self.enable_cancel_on_failover:
+            consumer_arguments = {
+                "x-cancel-on-ha-failover": True}
+
         self.queue = kombu.entity.Queue(
             name=self.queue_name,
             channel=conn.channel,
@@ -266,7 +281,9 @@ class Consumer(object):
             durable=self.durable,
             auto_delete=self.queue_auto_delete,
             routing_key=self.routing_key,
-            queue_arguments=self.queue_arguments)
+            queue_arguments=self.queue_arguments,
+            consumer_arguments=consumer_arguments
+        )
 
         try:
             LOG.debug('[%s] Queue.declare: %s',
@@ -467,6 +484,7 @@ class Connection(object):
         self.kombu_failover_strategy = driver_conf.kombu_failover_strategy
         self.kombu_compression = driver_conf.kombu_compression
         self.heartbeat_in_pthread = driver_conf.heartbeat_in_pthread
+        self.enable_cancel_on_failover = driver_conf.enable_cancel_on_failover
 
         if self.heartbeat_in_pthread:
             # NOTE(hberaud): Experimental: threading module is in use to run
@@ -524,7 +542,10 @@ class Connection(object):
                 self._url += url.virtual_host
         elif not url.hosts:
             host = oslo_messaging.transport.TransportHost('')
-            self._url = self._transform_transport_url(
+            # NOTE(moguimar): default_password in this function's context is
+            #                 a fallback option, not a hardcoded password.
+            #                 username and password are read from host.
+            self._url = self._transform_transport_url(  # nosec
                 url, host, default_username='guest', default_password='guest',
                 default_hostname='localhost')
 
@@ -654,7 +675,9 @@ class Connection(object):
         except KeyError:
             raise RuntimeError("Invalid SSL version : %s" % version)
 
-    def _transform_transport_url(self, url, host, default_username='',
+    # NOTE(moguimar): default_password in this function's context is just
+    #                 a fallback option, not a hardcoded password.
+    def _transform_transport_url(self, url, host, default_username='',  # nosec
                                  default_password='', default_hostname=''):
         transport = url.transport.replace('kombu+', '')
         transport = transport.replace('rabbit', 'amqp')
@@ -1003,7 +1026,8 @@ class Connection(object):
                     except (socket.timeout,
                             ConnectionRefusedError,
                             OSError,
-                            kombu.exceptions.OperationalError) as exc:
+                            kombu.exceptions.OperationalError,
+                            amqp_exec.ConnectionForced) as exc:
                         LOG.info("A recoverable connection/channel error "
                                  "occurred, trying to reconnect: %s", exc)
                         self.ensure_connection()
@@ -1109,31 +1133,35 @@ class Connection(object):
         responses for call/multicall
         """
 
-        consumer = Consumer(exchange_name='',  # using default exchange
-                            queue_name=topic,
-                            routing_key='',
-                            type='direct',
-                            durable=False,
-                            exchange_auto_delete=False,
-                            queue_auto_delete=False,
-                            callback=callback,
-                            rabbit_ha_queues=self.rabbit_ha_queues,
-                            rabbit_queue_ttl=self.rabbit_transient_queues_ttl)
+        consumer = Consumer(
+            exchange_name='',  # using default exchange
+            queue_name=topic,
+            routing_key='',
+            type='direct',
+            durable=False,
+            exchange_auto_delete=False,
+            queue_auto_delete=False,
+            callback=callback,
+            rabbit_ha_queues=self.rabbit_ha_queues,
+            rabbit_queue_ttl=self.rabbit_transient_queues_ttl,
+            enable_cancel_on_failover=self.enable_cancel_on_failover)
 
         self.declare_consumer(consumer)
 
     def declare_topic_consumer(self, exchange_name, topic, callback=None,
                                queue_name=None):
         """Create a 'topic' consumer."""
-        consumer = Consumer(exchange_name=exchange_name,
-                            queue_name=queue_name or topic,
-                            routing_key=topic,
-                            type='topic',
-                            durable=self.amqp_durable_queues,
-                            exchange_auto_delete=self.amqp_auto_delete,
-                            queue_auto_delete=self.amqp_auto_delete,
-                            callback=callback,
-                            rabbit_ha_queues=self.rabbit_ha_queues)
+        consumer = Consumer(
+            exchange_name=exchange_name,
+            queue_name=queue_name or topic,
+            routing_key=topic,
+            type='topic',
+            durable=self.amqp_durable_queues,
+            exchange_auto_delete=self.amqp_auto_delete,
+            queue_auto_delete=self.amqp_auto_delete,
+            callback=callback,
+            rabbit_ha_queues=self.rabbit_ha_queues,
+            enable_cancel_on_failover=self.enable_cancel_on_failover)
 
         self.declare_consumer(consumer)
 
@@ -1144,16 +1172,18 @@ class Connection(object):
         exchange_name = '%s_fanout' % topic
         queue_name = '%s_fanout_%s' % (topic, unique)
 
-        consumer = Consumer(exchange_name=exchange_name,
-                            queue_name=queue_name,
-                            routing_key=topic,
-                            type='fanout',
-                            durable=False,
-                            exchange_auto_delete=True,
-                            queue_auto_delete=False,
-                            callback=callback,
-                            rabbit_ha_queues=self.rabbit_ha_queues,
-                            rabbit_queue_ttl=self.rabbit_transient_queues_ttl)
+        consumer = Consumer(
+            exchange_name=exchange_name,
+            queue_name=queue_name,
+            routing_key=topic,
+            type='fanout',
+            durable=False,
+            exchange_auto_delete=True,
+            queue_auto_delete=False,
+            callback=callback,
+            rabbit_ha_queues=self.rabbit_ha_queues,
+            rabbit_queue_ttl=self.rabbit_transient_queues_ttl,
+            enable_cancel_on_failover=self.enable_cancel_on_failover)
 
         self.declare_consumer(consumer)
 
@@ -1359,6 +1389,10 @@ class RabbitDriver(amqpdriver.AMQPDriverBase):
         # the pool configuration properties
         max_size = conf.oslo_messaging_rabbit.rpc_conn_pool_size
         min_size = conf.oslo_messaging_rabbit.conn_pool_min_size
+        if max_size < min_size:
+            raise RuntimeError(
+                f"rpc_conn_pool_size: {max_size} must be greater than "
+                f"or equal to conn_pool_min_size: {min_size}")
         ttl = conf.oslo_messaging_rabbit.conn_pool_ttl
 
         connection_pool = pool.ConnectionPool(
